# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"UniqueNamer": "02_functions.ipynb",
         "FunctionsList": "02_functions.ipynb",
         "BaseFunction": "02_functions.ipynb",
         "Proportional": "02_functions.ipynb",
         "Variable": "02_functions.ipynb",
         "Subtract": "02_functions.ipynb",
         "Constant": "02_functions.ipynb",
         "Integration": "02_functions.ipynb",
         "WeightedSum": "02_functions.ipynb",
         "PCTNode": "03_nodes.ipynb",
         "PCTNodeData": "03_nodes.ipynb",
         "PCTHierarchy": "04_hierarchy.ipynb"}

modules = ["functions.py",
           "nodes.py",
           "hierarchy.py"]

doc_url = "https://perceptualrobots.github.io/pct/"

git_url = "https://github.com/perceptualrobots/pct/tree/master/"

def custom_doc_links(name): return None
